MCONFIG = {
    "application": {
        "name": "{0}",
        "author": "{1}",
        "version": "0.1",
    },
    "relations": [],
    "triggers": [],
    "trigger_params": {},
    "checkpoint": "0 * * * *",
}
